from __future__ import annotations

import itertools
import re
from datetime import date, timedelta
from typing import Optional, List, Tuple, Iterator

from .dates import at_end_of_month, read_date
from .ranges import range_incl


class DateRange:
    DAYS_PATTERN = re.compile(r'\s*DAYS\((\d+)\)\s*')

    @classmethod
    def maybe_timedelta(cls, value: str) -> Optional[timedelta]:
        m = cls.DAYS_PATTERN.match(value)
        if m:
            return timedelta(days=int(m.group(1)))
        return None

    @staticmethod
    def _date_or_begin_of_month(v: str) -> date:
        components = v.split('-')
        if len(components) == 2:
            year, month = components
            return date(int(year), int(month), 1)
        else:
            return read_date(v)

    @staticmethod
    def _date_or_end_of_month(v: str) -> date:
        components = v.split('-')
        if len(components) == 2:
            year, month = components
            d: date = date(int(year), int(month), 1)
            return at_end_of_month(d)
        else:
            return read_date(v)

    def __init__(self, date_range: str = ""):
        self._subranges: List[Tuple[Optional[date], Optional[date]]] = []
        if not date_range:
            return
        components = date_range.translate(str.maketrans('', '', ' ')).split(',')
        for component in components:
            range_components = component.split('..')
            if len(range_components) == 1:
                parts = component.split('-')
                if len(parts) == 2:
                    year, month = parts
                    year = int(year)
                    month = int(month)
                    lower: date = date(year, month, 1)
                    upper = at_end_of_month(lower)
                    self._subranges.append((lower, upper))
                else:
                    d = read_date(component)
                    self._subranges.append((d, d))
            elif len(range_components) == 2:
                c1, c2 = range_components
                d1 = self.maybe_timedelta(c1) or (c1 and self._date_or_begin_of_month(c1) or None)
                parser = self._date_or_begin_of_month if isinstance(d1, timedelta) else self._date_or_end_of_month
                d2 = self.maybe_timedelta(c2) or (c2 and parser(c2) or None)
                if isinstance(d1, timedelta):  # flip so time delta will always be d2
                    d1, d2 = d2, -d1
                if isinstance(d2, timedelta):
                    if isinstance(d1, timedelta) or d1 is None:
                        raise ValueError("date range cannot be composed only from time-deltas")
                    d1, d2 = sorted([d1 + d2, d1])
                self._subranges.append((d1, d2))

    def __contains__(self, item) -> bool:
        d = read_date(item)
        for d1, d2 in self._subranges:
            if d1 and d2:
                if d1 <= d <= d2:
                    return True
            elif d1:
                if d1 <= d:
                    return True
            else:
                if d <= d2:
                    return True
        return False

    def __eq__(self, other):
        if isinstance(other, date):
            return self._cmp(other) == 0
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, date):
            return self._cmp(other) <= 0
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, date):
            return self._cmp(other) < 0
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, date):
            return self._cmp(other) >= 0
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, date):
            return self._cmp(other) > 0
        return NotImplemented

    def _cmp(self, other: date):
        if (max_ := self.max) is not None and max_ < other:  # pylint: disable=E0601
            return -1
        if (min_ := self.min) is not None and min_ > other:  # pylint: disable=E0601
            return 1
        return 0

    @property
    def min(self) -> Optional[date]:
        if any(lower is None for lower, _ in self._subranges) or not self._subranges:
            return None
        return min(lower for lower, _ in self._subranges)

    @property
    def max(self) -> Optional[date]:
        if any(upper is None for _, upper in self._subranges) or not self._subranges:
            return None
        return max(upper for _, upper in self._subranges)

    def iter(self, lower=None, upper=None) -> Iterator[date]:
        lower = lower and read_date(lower) or self.min or None
        upper = upper and read_date(upper) or self.max or None

        if not lower or not upper:
            raise ValueError('no lower or upper bound')

        for d in range_incl(lower, upper):
            if d in self:
                yield d

    def itermonths(self) -> Iterator[DateRange]:
        for _key, group in itertools.groupby(iter(self), lambda d: d.month):
            yield DateRange(','.join(sorted(str(d) for d in group)))

    def __iter__(self) -> Iterator[date]:
        if not self._subranges:
            return
        min_ = self.min
        max_ = self.max
        if min_ is None or max_ is None:
            raise ValueError("open range")
        yield from self.iter(min_, max_)

    def _subranges_str(self) -> Iterator[str]:
        for lower, upper in self._subranges:
            if lower is None:
                yield f"..{upper}"
            elif upper is None:
                yield f"{lower}.."
            elif lower == upper:
                yield f"{lower}"
            else:
                yield f"{lower}..{upper}"

    def __str__(self) -> str:
        return ','.join(self._subranges_str())

    def __bool__(self):
        return bool(self._subranges)



import datetime
import unittest

from allthethings.ranges import range_incl, range_excl


class UuidTestCase(unittest.TestCase):

    def test_range_incl_with_dates(self):
        ds = list(range_incl(datetime.date(2020, 2, 27), datetime.date(2020, 3, 2)))
        self.assertEqual([
            datetime.date(2020, 2, 27),
            datetime.date(2020, 2, 28),
            datetime.date(2020, 2, 29),
            datetime.date(2020, 3, 1),
            datetime.date(2020, 3, 2),
        ], ds)

    def test_range_excl_with_dates(self):
        ds = list(range_excl(datetime.date(2020, 2, 27), datetime.date(2020, 3, 2)))
        self.assertEqual([
            datetime.date(2020, 2, 27),
            datetime.date(2020, 2, 28),
            datetime.date(2020, 2, 29),
            datetime.date(2020, 3, 1),
        ], ds)

    def test_range_incl_with_dates_step(self):
        ds = list(range_incl(datetime.date(2020, 2, 27), datetime.date(2020, 3, 2), step=2))
        self.assertEqual([
            datetime.date(2020, 2, 27),
            datetime.date(2020, 2, 29),
            datetime.date(2020, 3, 2),
        ], ds)

    def test_range_excl_with_dates_step(self):
        ds = list(range_excl(datetime.date(2020, 2, 27), datetime.date(2020, 3, 2), step=2))
        self.assertEqual([
            datetime.date(2020, 2, 27),
            datetime.date(2020, 2, 29),
        ], ds)


if __name__ == '__main__':
    unittest.main()



from typing import TypeVar, Dict, List, Iterable, Callable

U = TypeVar('U')
R = TypeVar('R')


def groupby(f: Callable[[U], R], xs: Iterable[U]) -> Dict[R, List[U]]:
    r = dict()
    for x in xs:
        k = f(x)
        if k not in r:
            r[k] = []
        r[k].append(x)
    return r



import datetime
import unittest
from typing import List

from allthethings.daterange import DateRange
from allthethings.dates import read_date


class TestDateRange(unittest.TestCase):

    def test_date_cmp(self):
        d1 = read_date('2020-04-01')
        d2 = read_date('2020-04-01')
        d3 = read_date('2020-04-02')
        self.assertTrue(d1 <= d2)
        self.assertFalse(d1 < d2)
        self.assertTrue(d1 < d3)

    def test_date_range_contains(self):
        d = DateRange('2020-04-01,2020-05-01..2020-05-20,2020-06-01..')
        self.assertTrue('2019-03-31' not in d)
        self.assertTrue('2020-03-31' not in d)
        self.assertTrue('2020-04-01' in d)
        self.assertTrue('2020-04-02' not in d)
        self.assertTrue('2020-05-01' in d)
        self.assertTrue('2020-05-10' in d)
        self.assertTrue('2020-05-20' in d)
        self.assertTrue('2020-05-21' not in d)
        self.assertTrue('2020-06-01' in d)
        self.assertTrue('2020-06-02' in d)
        self.assertTrue('2021-06-02' in d)

    def test_date_range_min_max(self):
        d = DateRange('2020-04-01,2020-05-01..2020-05-20,2020-06-01..')
        self.assertEqual(read_date('2020-04-01'), d.min)
        self.assertEqual(None, d.max)
        d = DateRange('2020-04-01,2020-05-01..2020-05-20,2020-06-01..,2019-01-01')
        self.assertEqual(read_date('2019-01-01'), d.min)
        self.assertEqual(None, d.max)
        d = DateRange('2020-04-01,2020-05-01..2020-05-20,2020-06-01..2020-06-15,2019-01-01')
        self.assertEqual(read_date('2019-01-01'), d.min)
        self.assertEqual(read_date('2020-06-15'), d.max)
        d = DateRange('2020-04-01,2020-05-01..2020-05-20,2031-03-15,2020-06-01..2020-06-15,..2019-01-01')
        self.assertEqual(None, d.min)
        self.assertEqual(read_date('2031-03-15'), d.max)

    def test_date_range_iter(self):
        d = DateRange('2020-01-01..2020-01-04')
        self.assertEqual([
            read_date('2020-01-01'),
            read_date('2020-01-02'),
            read_date('2020-01-03'),
            read_date('2020-01-04'),
        ], [*iter(d)])
        d = DateRange('2020-01-01..')
        self.assertEqual([
            read_date('2020-01-01'),
            read_date('2020-01-02'),
            read_date('2020-01-03'),
        ], [*d.iter(upper='2020-01-03')])

    def test_date_range_months(self):
        d = DateRange('2021-02')
        self.assertEqual(28, len([*d]))
        d = DateRange('2020-02')
        self.assertEqual(29, len([*d]))
        d = DateRange('2020-12')
        self.assertEqual(31, len([*d]))

    def test_date_range_month_bounds(self):
        d = DateRange('2020-01..2020-12')
        self.assertEqual(read_date('2020-01-01'), d.min)
        self.assertEqual(read_date('2020-12-31'), d.max)

    def test_date_range_itermonths(self):
        ranges: List[DateRange] = [*DateRange('2020-01..2020-12').itermonths()]
        self.assertEqual(12, len(ranges))
        for month, date_range in enumerate(ranges, start=1):
            self.assertEqual(1, date_range.min.day)
            self.assertEqual(month, date_range.min.month)
            self.assertEqual(month, date_range.max.month)

    def test_date_range_timedelta(self):
        def today():
            return datetime.datetime.utcnow().date()

        dr = DateRange('2020-01..DAYS(31)')
        self.assertEqual(dr.min, read_date('2020-01-01'))
        self.assertEqual(dr.max, read_date('2020-02-01'))
        dr = DateRange('DAYS(31)..2020-01')
        self.assertEqual(dr.min, read_date('2019-12-01'))
        self.assertEqual(dr.max, read_date('2020-01-01'))
        dr = DateRange('DAYS(31)..today')
        self.assertEqual(dr.min, read_date((today() - datetime.timedelta(days=31)).isoformat()))
        self.assertEqual(dr.max, read_date(today().isoformat()))
        dr = DateRange('today..DAYS(31)')
        self.assertEqual(dr.min, read_date(today().isoformat()))
        self.assertEqual(dr.max, read_date((today() + datetime.timedelta(days=31)).isoformat()))

    def test_date_comparison(self):
        r1 = DateRange('..2021-08-01')

        self.assertTrue(read_date('2021-07-01') <= r1)
        self.assertTrue(read_date('2021-07-01') >= r1)
        self.assertFalse(read_date('2021-07-01') < r1)
        self.assertFalse(read_date('2021-07-01') > r1)

        self.assertTrue(read_date('2021-08-01') <= r1)
        self.assertTrue(read_date('2021-08-01') >= r1)
        self.assertFalse(read_date('2021-08-01') < r1)
        self.assertFalse(read_date('2021-08-01') > r1)

        self.assertFalse(read_date('2021-08-02') <= r1)
        self.assertTrue(read_date('2021-08-02') >= r1)
        self.assertFalse(read_date('2021-08-02') < r1)
        self.assertTrue(read_date('2021-08-02') > r1)

        r2 = DateRange('2020-07-01..2020-08-01')

        self.assertTrue(read_date('2020-06-01') <= r2)
        self.assertFalse(read_date('2020-06-01') >= r2)
        self.assertTrue(read_date('2020-06-01') < r2)
        self.assertFalse(read_date('2020-06-01') > r2)

        self.assertTrue(read_date('2020-07-01') <= r2)
        self.assertTrue(read_date('2020-07-01') >= r2)
        self.assertFalse(read_date('2020-07-01') < r2)
        self.assertFalse(read_date('2020-07-01') > r2)

        self.assertTrue(read_date('2020-08-01') <= r2)
        self.assertTrue(read_date('2020-08-01') >= r2)
        self.assertFalse(read_date('2020-08-01') < r2)
        self.assertFalse(read_date('2020-08-01') > r2)

        self.assertFalse(read_date('2020-08-02') <= r2)
        self.assertTrue(read_date('2020-08-02') >= r2)
        self.assertFalse(read_date('2020-08-02') < r2)
        self.assertTrue(read_date('2020-08-02') > r2)

        r3 = DateRange('2019-08-01..')

        self.assertTrue(read_date('2019-07-01') <= r3)
        self.assertFalse(read_date('2019-07-01') >= r3)
        self.assertTrue(read_date('2019-07-01') < r3)
        self.assertFalse(read_date('2019-07-01') > r3)

        self.assertTrue(read_date('2019-08-01') <= r3)
        self.assertTrue(read_date('2019-08-01') >= r3)
        self.assertFalse(read_date('2019-08-01') < r3)
        self.assertFalse(read_date('2019-08-01') > r3)

        self.assertTrue(read_date('2019-08-02') <= r3)
        self.assertTrue(read_date('2019-08-02') >= r3)
        self.assertFalse(read_date('2019-08-02') < r3)
        self.assertFalse(read_date('2019-08-02') > r3)

    def test_empty_daterange(self):
        dr = DateRange()
        i = 0
        for d in dr:
            i += 1
        self.assertEqual(0, i)
        self.assertFalse(datetime.date(2020, 1, 1) in dr)

    def test_one_day_range(self):
        dr = DateRange("2020-01-01..2020-01-01")
        self.assertEqual([datetime.date(2020, 1, 1)], [*dr])


if __name__ == '__main__':
    unittest.main()



from datetime import date, datetime, timedelta
from typing import Union


def at_end_of_month(d: date) -> date:
    next_month = d.month % 12 + 1
    next_year = d.year + (1 if next_month == 1 else 0)
    return date(next_year, next_month, 1) - timedelta(days=1)


def read_date(value: Union[date, str]) -> date:
    if isinstance(value, date):
        return value
    if value.lower() == "today":
        return datetime.utcnow().date()
    if value.lower() == "yesterday":
        return datetime.utcnow().date() - timedelta(days=1)
    return date.fromisoformat(value)
